#
# Copyright 2012 - Max Hacking - Hacking Networked Solutions
#
# License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.

# Check if we have been included already
[[ -n "${HACKING_CERTIFICATE_LIB_LOADED+x}" ]] && return || HACKING_CERTIFICATE_LIB_LOADED="true"

function user_error
{
	echo "Error: ${1}" 1>&2
	exit 1
}

function internal_error
{
	echo "Internal error: ${1}" 1>&2
	exit 2
}

function ensure_directory_empty
{
	# Check that the current directory is empty
	[[ "$(ls -A)" ]] && user_error "${cmdname} should only be run in ${1} mode in an empty directory"
}

function assemble_configuration
{
	# Check that we were at least passed something to work with.
	(( $# == 0 )) && internal_error "assemble_configuration() called with no configuration files!"
	
	# Delete any configuration file which already exits.
	rm -f "${ossl_config_file}"
	
	# Loop through the parameters, which should be paths to config files.
	while (( $# > 0 ))
	do	
		# If this is not a path to a readable file then complain and quit.
		[[ ! -r "${1}" ]] &&  internal_error "assemble_configuration() called with unreadable or nonexistant file [${1}]"

		# Append the source configuration (minus ## quoted lines) to the temporary configuration file
		echo -e "## Configuration read from file ${1} starts here\n" >> "${ossl_config_file}"
		sed '/^\##/d' "${1}" >> "${ossl_config_file}"
						
		# Move on to the next parameter
		shift 1
	done
	
	# If we are running in verbose mode then display the generated configuration
	if (( ${verbose} ))
	then
		echo -e "Assembled configuration file\n"
		cat "${ossl_config_file}"
		echo
	fi
}

function is_power_of_two
{	
	(( ${1} == 0 )) && return 1
	
	local number=${1}
 	while (( $((number % 2)) == 0 ))
 	do
 		number=$((${number} / 2))
	done
	
	(( ${number} > 1 )) && return 1
	
	return 0
}

function validate_key_parameters
{	
	# Convert algorithm to upper case.
	algorithm="${algorithm^^}"
	
	# If we were supplied no algorithm then set RSA as the default.
	[[ -z "${algorithm}" ]] && algorithm="RSA"
	
	# Attempt to validate the algorithm and options.
	case "${algorithm}" in
		# For RSA, DSA or DH the options should contain a key length.
		RSA|DSA|DH)
			if [[ -z "${algopts}" ]]
			then
				algopts="4096"
				return
			fi
			[[ ! "${algopts}" =~ ^[0-9]+$ ]] && user_error "The key length should be an integer"
			is_power_of_two ${algopts} || user_error "The key length should be a power of two"
			(( ${algopts} < 512 )) && user_error "Key length must be > 512"		
			(( ${algopts} > 8192 )) && user_error "Key length must be < 8192"		
		;;
		# For EC the options should contain a valid curve name.
		ECDSA)
			if [[ -z "${algopts}" ]]
			then
				algopts="secp384r1"
				return
			fi
			openssl ecparam -list_curves | grep "[[:space:]]${algopts}[[:space:]]" > /dev/null
			(( $? )) && user_error "Invalid curve name [${algopts}]"
		;;
		# If we got this far then we were called with an unknown algorith
		*)
			user_error "Unknown algorithm [${algorithm}]"
		;;
	esac		
}

function create_key_file
{
	local retval=0
	local options
	
	# Check we were called with usable parameters
	[[ -z "${1}" ]] && internal_error "create_key_file() called with no file name!"
	[[ -z "${2}" ]] && internal_error "create_key_file() called with no algorithm!"
		
	# Generate the appropriate type of key. 
	case "${2}" in
		RSA)
			[[ -n "${3}" ]] && options="-pkeyopt rsa_keygen_bits:${3}"
			openssl genpkey -algorithm RSA ${options} -out "${1}"
			retval=$?
		;;
		DSA)
			[[ -n "${3}" ]] && options="-pkeyopt dsa_paramgen_bits:${3}"
			openssl genpkey -algorithm DSA ${options} -out "${1}"
			retval=$?
		;;
		DH)
			[[ -n "${3}" ]] && options="-pkeyopt dh_paramgen_prime_len:${3}"
			openssl genpkey -algorithm DH ${options} -out "${1}"
			retval=$?
		;;
		ECDSA)
			[[ -z "${3}" ]] && internal_error "create_key_file() called with algorithm set to ECDSA and no curve name!"
			openssl ecparam -genkey -name ${3} -out "${1}"
			retval=$?
		;;
		*)
			internal_error "create_key_file() called with unknown algorithm ${2}"
		;;
	esac
	
	# Check that we created a key file, if not then bail.
	(( ${retval} )) && internal_error "Unable to create key file"
}

function encrypt_key_file
{
	# Check we were called with usable parameters
	[[ -z "${1}" ]] && internal_error "encrypt_key_file() called with no file name!"
	[[ -z "${1}" ]] && internal_error "encrypt_key_file() called with non-existant file!"

	# Encrypt the key to a new file. 
	openssl pkey -in "${1}" -des3 -out "${1}_tmp"
	
	# If it worked then swap the files, otherwise complain and quit.
	[[ -f "${1}_tmp" ]] && mv "${1}_tmp" "${1}" \
						|| internal_error "Failed to encrypt key file."
}

function get_dnfn_from_request
{
	# Check we were called with a file name and that the file exists
	[[   -z "${1}" ]] && internal_error "get_dnfn_from_request() called with no file name!"
	[[ ! -f "${1}" ]] && internal_error "get_dnfn_from_request() called with non-existant file ${1}"

    # Get the common name from the new certificate request
	local req_dn=$(openssl req -in "${1}" -subject -noout | \
				   sed -e 's/subject=//' -e 's,/[^=]*=, ,g' -e 's, *,,' -e 's/ \| - /_/g')

	# Check it worked and bail if not.
	[[ -z "${req_dn}" ]] && internal_error "get_dnfn_from_request() unable to get DN"

	# Create a filename from the common name
    dnfn="${req_dn// - /_}"
    dnfn="${dnfn//\'/}"
    dnfn="${dnfn// /_}"
}

function get_dnfn_from_certificate
{
	# Check we were called with a file name and that the file exists
	[[   -z "${1}" ]] && internal_error "get_dnfn_from_certificate() called with no file name!"
	[[ ! -f "${1}" ]] && internal_error "get_dnfn_from_certificate() called with non-existant file ${1}"

    # Get the distinguished name from the new certificate
	local cert_dn=$(openssl x509 -in "${1}" -subject -noout | \
					sed -e 's/subject=//' -e 's,/[^=]*=, ,g' -e 's, *,,' -e 's/ \| - /_/g')

	# Check it worked and bail if not.
	[[ -z "${cert_dn}" ]] && internal_error "get_dnfn_from_certificate() unable to get DN"

	# Create a filename from the common name
    dnfn="${cert_dn// - /_}"
    dnfn="${dnfn//\'/}"
    dnfn="${dnfn// /_}"
}

function get_skid_from_request
{
	# Check we were called with a file name and that the file exists
	[[   -z "${1}" ]] && internal_error "get_skid_from_request() called with no file name!"
	[[ ! -f "${1}" ]] && internal_error "get_skid_from_request() called with non-existant file ${1}"
	
	# Get the Subject Key Identifier and strip the whitespace and : chars.	
	skid=$(openssl req -noout -text -in "${1}" | grep "Subject Key Identifier:" --after-context=1 | tail -n 1)
	skid="${skid//:/}"
	skid="${skid//[[:space:]]/}"
		
	# Whinge and quit if we didn't manage to find it.
	[[ -z "${skid}" ]] && internal_error "get_skid_from_request() - unable to get Subject Key Identifier!"
}

function get_skid_from_certificate
{
	# Check we were called with a file name and that the file exists
	[[   -z "${1}" ]] && internal_error "get_skid_from_certificate() called with no file name!"
	[[ ! -f "${1}" ]] && internal_error "get_skid_from_certificate() called with non-existant file ${1}"
	
	# Get the Subject Key Identifier and strip the whitespace and : chars.	
	skid=$(openssl x509 -noout -text -in "${1}" | grep "Subject Key Identifier:" --after-context=1 | tail -n 1)
	skid="${skid//:/}"
	skid="${skid//[[:space:]]/}"
		
	# Whinge and quit if we didn't manage to find it.
	[[ -z "${skid}" ]] && internal_error "get_skid_from_certificate() - unable to get Subject Key Identifier!"
}

function match_keys
{
	# Check we were called with a file name and that the file exists
	[[   -z "${1}" ]] && internal_error "match_keys() called with no certificate directory!"
	[[ ! -d "${1}" ]] && internal_error "match_keys() called with non-existant certificate directory ${1}"
	[[   -z "${2}" ]] && internal_error "match_keys() called with no key source directory!"
	[[ ! -d "${2}" ]] && internal_error "match_keys() called with non-existant key source directory ${2}"
	[[   -z "${3}" ]] && internal_error "match_keys() called with no key destination directory!"
	[[ ! -d "${3}" ]] && internal_error "match_keys() called with non-existant key destination directory ${3}"
				
	# Loop through all the certificate files in the auth_cert_dir
	local cert_file_list=$( find ${1}/ -name "*.crt.pem" )
	for cert_file in ${cert_file_list}
	do
		# Does this certificate file already have a matching key?
		local key_file="${cert_file//\.crt\.pem/}.key.pem"
		key_file="${key_file/$1/$3}"
		if [[ ! -f "${key_file}" ]]
		then
			# It doesn't, let's get its Subject Key Identifier and see if we can find it.
			local skid
			get_skid_from_certificate "${cert_file}"

			# If a key file named with this SKID exists copy it to the auth_keys_dir,
			# complain about lack of matching key if not.
			if [[ -f "${2}/${skid}.key.pem" ]]
			then
				echo "Found key for ${cert_file##*/} : ${skid}.key.pem"
				cp "${2}/${skid}.key.pem" "${key_file}"
			else
				echo "A matching key for file ${cert_file##*/} could not be found."
				echo "Expected to find file ${2}/${skid}.key.pem"
			fi
		fi
	done
}

